from __future__ import annotations
from contextlib import contextmanager
from typing import (
    TYPE_CHECKING,
    List,
    Optional,
)

from signe.core.protocols import IScope
from weakref import WeakSet

if TYPE_CHECKING:  # pragma: no cover
    from .protocols import DisposableProtocol


class Scope(IScope):
    def __init__(self) -> None:
        self._disposables: WeakSet[DisposableProtocol] = WeakSet()

    def add_disposable(self, disposable: DisposableProtocol):
        self._disposables.add(disposable)

    def dispose(self):
        for effect in self._disposables:
            effect.dispose()

        self._disposables.clear()


class GlobalScopeManager:
    def __init__(self) -> None:
        self._stack: List[Scope] = []

    def _get_last_scope(self):
        idx = len(self._stack) - 1
        if idx < 0:
            return None

        return self._stack[idx]

    def new_scope(self):
        s = Scope()
        self._stack.append(s)
        return s

    def dispose_scope(self):
        s = self._get_last_scope()
        if s:
            s.dispose()
            self._stack.pop()

    def mark_effect_with_scope(
        self, scope: Optional[Scope], effect: DisposableProtocol
    ):
        s = scope
        if s:
            s.add_disposable(effect)

        return effect

    def mark_effect(self, effect: DisposableProtocol):
        return self.mark_effect_with_scope(self._get_last_scope(), effect)


_GLOBAL_SCOPE_MANAGER = GlobalScopeManager()


@contextmanager
def scope():
    scope = _GLOBAL_SCOPE_MANAGER.new_scope()
    yield scope
    _GLOBAL_SCOPE_MANAGER.dispose_scope()



from datetime import datetime, date
import inspect
from typing import Callable, Tuple


def get_func_args_count(fn):
    return len(inspect.getfullargspec(fn).args)


def is_object(obj):
    return obj is not None and (
        not isinstance(obj, (int, str, Tuple, float, datetime, date, Callable))
    )


def has_changed(a, b):
    try:
        return bool(a != b)
    except ValueError:  # pragma: no cover
        return True



from __future__ import annotations
from typing import Callable, Dict


from .collections import Stack
from .protocols import CallerProtocol


# def _defatul_executor_builder():
#     return Executor()  # pragma: no cover


# executor_builder = _defatul_executor_builder


# class Executor:
#     def __init__(self) -> None:
#         self._caller_running_stack = Stack[CallerProtocol]()
#         self.execution_scheduler_stack = Stack[ExecutionScheduler]()
#         self.__defalut_executionScheduler = ExecutionScheduler()
#         self._pause_track_count = 0

#     def pause_track(self):
#         self._pause_track_count += 1  # pragma: no cover

#     def reset_track(self):
#         self._pause_track_count -= 1  # pragma: no cover

#     def should_track(self):
#         return self._pause_track_count <= 0

#     def set_default_execution_scheduler(self, execution_scheduler: ExecutionScheduler):
#         self.__defalut_executionScheduler = execution_scheduler  # pragma: no cover

#     def get_current_scheduler(self):
#         return (
#             self.execution_scheduler_stack.get_current()
#             or self.__defalut_executionScheduler
#         )

#     def mark_running_caller(self, caller: CallerProtocol):
#         self._caller_running_stack.set_current(caller)

#     def reset_running_caller(self, caller: CallerProtocol):
#         self._caller_running_stack.reset_current()

#     def get_running_caller(self):
#         return self._caller_running_stack.get_current()


class ExecutionScheduler:
    def __init__(self) -> None:
        self._caller_running_stack = Stack[CallerProtocol]()
        self._pause_track_count = 0

        self._scheduler_fns: Dict[Callable[[], None], None] = {}
        self.__running = False
        self.pause_should_run_stack = 0

    def pause_track(self):
        self._pause_track_count += 1  # pragma: no cover

    def reset_track(self):
        self._pause_track_count -= 1  # pragma: no cover

    def should_track(self):
        return self._pause_track_count <= 0

    def mark_running_caller(self, caller: CallerProtocol):
        self._caller_running_stack.set_current(caller)

    def reset_running_caller(self, caller: CallerProtocol):
        self._caller_running_stack.reset_current()

    def get_running_caller(self):
        return self._caller_running_stack.get_current()

    @property
    def should_run(self):
        return not (self.__running or (self.pause_should_run_stack != 0))

    def pause_scheduling(self):
        self.pause_should_run_stack += 1

    def reset_scheduling(self):
        self.pause_should_run_stack -= 1

    def push_scheduler_fn(self, fn: Callable[[], None]):
        self._scheduler_fns[fn] = None

    def run(self):
        count = 0
        self.__running = True

        try:
            while self._scheduler_fns:
                self._run_scheduler_fns()

                count += 1
                if count >= 10000:  # pragma: no cover
                    raise Exception("exceeded the maximum number of execution rounds.")
        finally:
            self.__running = False

    def _run_scheduler_fns(self):
        fns = tuple(self._scheduler_fns.keys())
        self._scheduler_fns.clear()
        for fn in fns:
            fn()


# class BatchExecutionScheduler(ExecutionScheduler):
#     def __init__(self) -> None:
#         super().__init__()

#     @property
#     def should_run(self):
#         return False

#     def run(self):  # pragma: no cover
#         pass

#     def run_batch(self):
#         return super().run()



from __future__ import annotations
from collections import UserDict, UserList
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Iterator,
    List,
    Mapping,
    Optional,
    TypeVar,
    cast,
    Iterable,
)
from signe.core.consts import EffectState
from signe.core.context import get_default_scheduler
from signe.core.deps import GetterDepManager
from signe.core.helper import has_changed, is_object
from signe.core.protocols import RawableProtocol
from .batch import batch
from weakref import WeakKeyDictionary, WeakValueDictionary


if TYPE_CHECKING:  # pragma: no cover
    from .runtime import ExecutionScheduler


T = TypeVar("T")
P = TypeVar("P")


_proxy_maps: WeakValueDictionary = WeakValueDictionary()


def track_all_deep(obj):
    stack = [obj]

    while len(stack):
        current = stack.pop()
        if isinstance(current, ListProxy):
            for value in current:
                if is_reactive(value):
                    stack.append(value)

        elif isinstance(current, DictProxy):
            for value in current.values():
                if is_reactive(value):
                    stack.append(value)
        elif isinstance(current, InstanceProxy):
            for key in _get_data_fields(current):
                value = getattr(current, key)
                if is_reactive(value):
                    stack.append(value)
        else:
            pass  # pragma: no cover


def track_all(obj, scheduler: ExecutionScheduler, deep=False):
    if not scheduler.should_track():
        return  # pragma: no cover

    if isinstance(obj, (DictProxy, ListProxy)):
        if deep:
            track_all_deep(obj)
        else:
            iter(obj)
    elif isinstance(obj, (InstanceProxy)):
        if deep:
            track_all_deep(obj)
        else:
            for key in _get_data_fields(obj):
                getattr(obj, key)
    else:
        pass  # pragma: no cover


def reactive(
    obj: T,
    scheduler: Optional[ExecutionScheduler] = None,
) -> T:
    if not is_object(obj) or is_reactive(obj):
        return cast(T, obj)

    obj_id = id(obj)
    proxy = _proxy_maps.get(obj_id)

    if proxy:
        return proxy  # type: ignore

    scheduler = scheduler or get_default_scheduler()

    if isinstance(obj, Mapping):
        proxy = DictProxy(obj, scheduler)

    elif isinstance(obj, List):
        proxy = ListProxy(obj, scheduler)
    else:
        proxy = InstanceProxy(obj, scheduler)

    _proxy_maps[obj_id] = proxy
    return cast(T, proxy)

    # return obj


def to_reactive(obj: T, scheduler: ExecutionScheduler) -> T:
    res = reactive(obj, scheduler) if is_object(obj) else obj
    return cast(T, res)


def is_reactive(obj: object) -> bool:
    return _is_proxy(obj)


def to_raw(obj: T) -> T:
    if isinstance(obj, RawableProtocol):
        return obj.to_raw()

    if isinstance(obj, InstanceProxy):
        return _instance_proxy_maps.get(obj)
    return obj


def _is_proxy(obj):
    return isinstance(obj, (DictProxy, ListProxy, InstanceProxy))


class DictProxy(UserDict):
    def __init__(
        self,
        data,
        scheduler: ExecutionScheduler,
    ):
        super().__init__()
        self.data = data
        self._dep_manager = GetterDepManager(scheduler)
        self.__nested = set()
        self._scheduler = scheduler

    def __getitem__(self, key):
        self._dep_manager.tracked(key)
        res = reactive(self.data[key], self._scheduler)
        if _is_proxy(res):
            self.__nested.add(res)
        return res

    def __setitem__(self, key, item):
        item = to_raw(item)
        is_new = key not in self.data

        @batch
        def _():
            if is_new:
                self.data[key] = item
                self._dep_manager.triggered(
                    "len", len(self.data), EffectState.NEED_UPDATE
                )

            else:
                org_value = self.data[key]
                self.data[key] = item

                if has_changed(org_value, item):
                    self._dep_manager.triggered(key, item, EffectState.NEED_UPDATE)

            self._dep_manager.triggered("__iter__", None, EffectState.NEED_UPDATE)

    def __iter__(self) -> Iterator:
        self._dep_manager.tracked("__iter__")
        return super().__iter__()

    def __len__(self) -> int:
        self._dep_manager.tracked("len")

        return len(self.data)

    def __contains__(self, key: object) -> bool:
        result = key in self.data
        self._dep_manager.tracked("len")
        return result

    def __delitem__(self, key):
        del self.data[key]
        self._dep_manager.triggered("len", len(self.data), EffectState.NEED_UPDATE)

    def pop(self, i: int = -1) -> Any:
        return super().pop(i)

    def clear(self) -> None:
        super().clear()

    def __hash__(self) -> int:
        return hash(id(self))

    def __eq__(self, __value: object) -> bool:  # pragma: no cover
        if isinstance(__value, self.__class__):
            return self.__hash__() == __value.__hash__()

        return False

    def to_raw(self):
        return self.data


class ListProxy(UserList):
    def __init__(
        self,
        initlist,
        scheduler: ExecutionScheduler,
    ):
        super().__init__()
        self.data = initlist
        self._dep_manager = GetterDepManager(scheduler)
        self.__nested = set()
        self._scheduler = scheduler

    def __getitem__(self, i):
        self._dep_manager.tracked(i)
        res = reactive(self.data[i], self._scheduler)
        if _is_proxy(res):
            self.__nested.add(res)
        return res

    def __setitem__(self, i, item):
        item = to_raw(item)
        org_value = self.data[i]
        self.data[i] = item

        if has_changed(org_value, item):

            @batch
            def _():
                self._dep_manager.triggered(i, item, EffectState.NEED_UPDATE)
                self._dep_manager.triggered("__iter__", None, EffectState.NEED_UPDATE)

    def __iter__(self) -> Iterator:
        self._dep_manager.tracked("__iter__")
        return super().__iter__()

    def __len__(self) -> int:
        self._dep_manager.tracked("len")

        return len(self.data)

    def append(self, item: Any) -> None:
        super().append(to_raw(item))

        @batch
        def _():
            self._dep_manager.triggered("len", len(self.data), EffectState.NEED_UPDATE)
            self._dep_manager.triggered("__iter__", None, EffectState.NEED_UPDATE)

    def extend(self, other: Iterable) -> None:
        super().extend((to_raw(o) for o in other))

        @batch
        def _():
            self._dep_manager.triggered("len", len(self.data), EffectState.NEED_UPDATE)
            self._dep_manager.triggered("__iter__", None, EffectState.NEED_UPDATE)

    def remove(self, item: Any) -> None:
        super().remove(to_raw(item))

        @batch
        def _():
            self._dep_manager.triggered("len", len(self.data), EffectState.NEED_UPDATE)
            self._dep_manager.triggered("__iter__", None, EffectState.NEED_UPDATE)

    def pop(self, i: int = -1) -> Any:
        value = super().pop(i)

        @batch
        def _():
            self._dep_manager.triggered("len", len(self.data), EffectState.NEED_UPDATE)
            self._dep_manager.triggered("__iter__", None, EffectState.NEED_UPDATE)

        return value

    def clear(self) -> None:
        super().clear()

        @batch
        def _():
            self._dep_manager.triggered("len", len(self.data), EffectState.NEED_UPDATE)
            self._dep_manager.triggered("__iter__", None, EffectState.NEED_UPDATE)

    def __contains__(self, item) -> bool:
        result = to_raw(item) in self.data
        self._dep_manager.tracked("len")
        return result

    def __str__(self) -> str:
        return str(self.data)

    def __hash__(self) -> int:
        return hash(id(self))

    def __eq__(self, __value: object) -> bool:
        if isinstance(__value, self.__class__):
            return self.__hash__() == __value.__hash__()

        return False

    def to_raw(self):
        return self.data


_instance_proxy_maps: WeakKeyDictionary = WeakKeyDictionary()
_instance_dep_maps: WeakKeyDictionary = WeakKeyDictionary()
_instance_nested: WeakKeyDictionary = WeakKeyDictionary()


def _register_ins(
    proxy: InstanceProxy,
    ins,
    scheduler: ExecutionScheduler,
):
    _instance_proxy_maps[proxy] = ins
    _instance_dep_maps[proxy] = GetterDepManager(scheduler)


def _is_instance_method(obj, key: str):
    return isinstance(getattr(obj, key), Callable)


def _track_ins(
    proxy: InstanceProxy,
    key,
):
    ins = _instance_proxy_maps.get(proxy)
    if _is_instance_method(ins, key):
        method = getattr(ins, key)
        return method

        # def replace_method(instance, *args, **kws):
        #     method.__func__(instance, *args, **kws)
        #     method(*args, **kws)

        # fake_method = types.MethodType(replace_method, proxy)
        # return fake_method
    else:
        dep_manager = _instance_dep_maps.get(proxy)
        # assert ins
        assert dep_manager

        dep_manager.tracked(key)

        value = reactive(getattr(ins, key), dep_manager._scheduler)
        if _is_proxy(value):
            _instance_nested[proxy] = value
        return value


def _trigger_ins(proxy: InstanceProxy, key, value):
    ins = _instance_proxy_maps.get(proxy)
    dep_manager = _instance_dep_maps.get(proxy)

    assert ins
    assert dep_manager

    setattr(ins, key, value)
    dep_manager.triggered(key, value, EffectState.NEED_UPDATE)


def _get_data_fields(proxy: InstanceProxy):
    ins = _instance_proxy_maps.get(proxy)
    return [f for f in dir(ins) if f[0] != "_"]


class InstanceProxy:
    def __init__(
        self,
        ins,
        scheduler: ExecutionScheduler,
    ) -> None:
        _register_ins(self, ins, scheduler)

    def __getattr__(self, _name: str) -> Any:
        value = _track_ins(self, _name)
        return value

    def __setattr__(self, _name: str, _value: Any) -> None:
        _trigger_ins(self, _name, _value)



from signe.core.signal import Signal, signal, to_value, is_signal
from signe.core.effect import Effect, effect, stop
from signe.core.computed import Computed, computed
from signe.core.runtime import ExecutionScheduler
from signe.core.batch import batch
from signe.core.on import on, WatchedState
from signe.core.cleanup import cleanup
from signe.core.reactive import reactive, to_raw, is_reactive
from signe.core.scope import scope
from signe.core.types import TMaybeSignal, TGetterSignal, TSignal, TGetter

__all__ = [
    "Signal",
    "Effect",
    "Computed",
    "ExecutionScheduler",
    "signal",
    "effect",
    "computed",
    "batch",
    "on",
    "to_value",
    "is_signal",
    "cleanup",
    "reactive",
    "scope",
    "stop",
    "to_raw",
    "TMaybeSignal",
    "TGetterSignal",
    "TGetter",
    "TSignal",
    "WatchedState",
    "is_reactive",
]

__version__ = "0.4.6"



from typing import Callable, TypeVar, Union
from signe.core.protocols import ComputedResultProtocol, SignalResultProtocol


_T = TypeVar("_T")

TSignal = SignalResultProtocol[_T]
TGetterSignal = Union[SignalResultProtocol[_T], ComputedResultProtocol[_T]]
TGetter = Union[TGetterSignal[_T], Callable[[], _T]]
TMaybeSignal = Union[_T, TGetter[_T]]



from typing import (
    Callable,
    Optional,
)

from signe.core.context import get_default_scheduler

from .runtime import ExecutionScheduler


def cleanup(
    fn: Callable[[], None],
    scheduler: Optional[ExecutionScheduler] = None,
):
    return api_cleanup(fn, scheduler or get_default_scheduler())


def api_cleanup(
    fn: Callable[[], None],
    scheduler: ExecutionScheduler,
):
    current_caller = scheduler.get_running_caller()
    if current_caller:
        current_caller.add_cleanup(fn)



from typing import (
    Callable,
    Optional,
)

from signe.core.context import get_default_scheduler

from .runtime import ExecutionScheduler


def batch(
    fn: Callable[[], None],
    scheduler: Optional[ExecutionScheduler] = None,
):
    return api_batch(fn, scheduler or get_default_scheduler())


def api_batch(
    fn: Callable[[], None],
    scheduler: ExecutionScheduler,
):
    scheduler.pause_scheduling()

    try:
        fn()
    finally:
        scheduler.reset_scheduling()
        scheduler.run()



from __future__ import annotations
from typing import (
    TYPE_CHECKING,
    TypeVar,
    Generic,
    Callable,
    Union,
    Optional,
    cast,
    overload,
)
from signe.core.computed import Computed
from signe.core.reactive import to_raw, to_reactive
from signe.core.consts import EffectState
from signe.core.idGenerator import IdGen

from signe.core.deps import GetterDepManager
from signe.core.protocols import SignalResultProtocol
from .context import get_default_scheduler
from .types import TMaybeSignal, TSignal
from collections.abc import Hashable
import operator


if TYPE_CHECKING:  # pragma: no cover
    from .runtime import ExecutionScheduler

_T = TypeVar("_T")

TComp = TypeVar("TComp")

TSignalOptionComp = Callable[[TComp, TComp], bool]
TSignalOptionInitComp = Optional[Union[bool, TSignalOptionComp[TComp]]]


def _eq_base_comp(old, new):
    if isinstance(old, Hashable) and isinstance(new, Hashable):
        return operator.eq(old, new)

    return old is new


def _always_false_comp(old, new):
    return False


class SignalOption(Generic[_T]):
    __slots__ = ("comp",)

    def __init__(self, comp: TSignalOptionInitComp[_T] = None) -> None:
        if comp is None:
            comp = _eq_base_comp
        elif comp is False:
            comp = _always_false_comp
        self.comp: TSignalOptionComp = comp  # type: ignore


class Signal(Generic[_T]):
    __slots__ = (
        "__id",
        "_is_shallow",
        "_value",
        "_raw_value",
        "_scheduler",
        "_dep_manager",
        "option",
        "__debug_name",
        "_option_comp",
        "__weakref__",
    )
    _id_gen = IdGen("Signal")

    def __init__(
        self,
        value: _T,
        *,
        scheduler: ExecutionScheduler,
        option: Optional[SignalOption[_T]] = None,
        debug_name: Optional[str] = None,
        is_shallow: bool,
    ) -> None:
        super().__init__()
        self.__id = Signal._id_gen.new()
        self._scheduler = scheduler or get_default_scheduler()
        self._is_shallow = is_shallow
        self._value = value if is_shallow else to_reactive(value, self._scheduler)
        self._raw_value = value if is_shallow else to_raw(value)

        self._dep_manager = GetterDepManager(self._scheduler)

        self.option = option or SignalOption[_T]()
        self.__debug_name = debug_name
        self._option_comp = cast(Callable[[_T, _T], bool], self.option.comp)

    @property
    def id(self):
        return self.__id  # pragma: no cover

    @property
    def value(self):
        self._dep_manager.tracked("value")

        return self._value

    @value.setter
    def value(self, value: _T):
        use_direct = self._is_shallow
        new_value = value if use_direct else to_raw(value)

        if self._option_comp(self._raw_value, new_value):  # type: ignore
            return

        self._raw_value = new_value
        self._value = (
            new_value if use_direct else to_reactive(new_value, self._scheduler)
        )

        self._dep_manager.triggered("value", new_value, EffectState.NEED_UPDATE)

    def __repr__(self) -> str:
        return f"Signal(id= {self.id} , name = {self.__debug_name})"


@overload
def signal(
    value: SignalResultProtocol[_T],
    comp: Union[TSignalOptionInitComp[_T], bool] = None,
    debug_name: Optional[str] = None,
    *,
    scheduler: Optional[ExecutionScheduler] = None,
) -> SignalResultProtocol[_T]:
    ...


@overload
def signal(
    value: _T,
    comp: Union[TSignalOptionInitComp[_T], bool] = None,
    debug_name: Optional[str] = None,
    *,
    is_shallow=False,
    scheduler: Optional[ExecutionScheduler] = None,
) -> SignalResultProtocol[_T]:
    ...


@overload
def signal(
    value: TMaybeSignal[_T],
    comp: Union[TSignalOptionInitComp[_T], bool] = None,
    debug_name: Optional[str] = None,
    *,
    is_shallow=False,
    scheduler: Optional[ExecutionScheduler] = None,
) -> SignalResultProtocol[_T]:
    ...


def signal(
    value: Union[_T, SignalResultProtocol[_T], TMaybeSignal[_T]],
    comp: Union[TSignalOptionInitComp[_T], bool] = None,
    debug_name: Optional[str] = None,
    *,
    is_shallow=False,
    scheduler: Optional[ExecutionScheduler] = None,
) -> SignalResultProtocol[_T]:
    if isinstance(value, Signal):
        return value
    signal = Signal(
        value,
        scheduler=scheduler or get_default_scheduler(),
        option=SignalOption(comp),
        debug_name=debug_name,
        is_shallow=is_shallow,
    )
    return cast(SignalResultProtocol[_T], signal)


def is_signal(obj: TMaybeSignal):
    """Checks if a value is a signal or computed object.

    Args:
        obj (_type_): _description_
    """
    return isinstance(obj, (Signal, Computed))


def to_value(obj: TMaybeSignal[_T]) -> _T:
    """Normalizes values / signals / getters to values.

    Args:
        obj (_type_): _description_

    ## Example
    ```
    to_value(1)          #    --> 1
    to_value(signal(1))  #    --> 1
    ```

    """
    if is_signal(obj):
        return cast(TSignal, obj).value

    # if isinstance(obj, Callable):
    #     return obj()

    return cast(_T, obj)
